import base64
import functools
import re
import shutil
import subprocess
from pathlib import Path

import ciso8601
import requests
from mutagen.mp4 import MP4, MP4Cover
from pywidevine import PSSH, Cdm, Device
from yt_dlp import YoutubeDL

from .apple_music_api import AppleMusicApi
from .constants import MP4_TAGS_MAP
from .enums import CoverFormat, DownloadMode, RemuxMode
from .hardcoded_wvd import HARDCODED_WVD
from .itunes_api import ItunesApi
from .models import DownloadQueueItem, UrlInfo


class Downloader:
    ILLEGAL_CHARACTERS_REGEX = r'[\\/:*?"<>|;]'

    def __init__(
        self,
        apple_music_api: AppleMusicApi,
        itunes_api: ItunesApi,
        output_path: Path = Path("./Apple Music"),
        temp_path: Path = Path("./temp"),
        wvd_path: Path = None,
        nm3u8dlre_path: str = "N_m3u8dl-RE",
        mp4decrypt_path: str = "mp4decrypt",
        ffmpeg_path: str = "ffmpeg",
        mp4box_path: str = "MP4Box",
        download_mode: DownloadMode = DownloadMode.YTDLP,
        remux_mode: RemuxMode = RemuxMode.FFMPEG,
        cover_format: CoverFormat = CoverFormat.JPG,
        template_folder_album: str = "{album_artist}/{album}",
        template_folder_compilation: str = "Compilations/{album}",
        template_file_single_disc: str = "{track:02d} {title}",
        template_file_multi_disc: str = "{disc}-{track:02d} {title}",
        template_folder_no_album: str = "{artist}/Unknown Album",
        template_file_no_album: str = "{title}",
        template_date: str = "%Y-%m-%dT%H:%M:%SZ",
        exclude_tags: str = None,
        cover_size: int = 1200,
        truncate: int = 40,
        no_progress: bool = False,
    ):
        self.apple_music_api = apple_music_api
        self.itunes_api = itunes_api
        self.output_path = output_path
        self.temp_path = temp_path
        self.wvd_path = wvd_path
        self.nm3u8dlre_path = nm3u8dlre_path
        self.mp4decrypt_path = mp4decrypt_path
        self.ffmpeg_path = ffmpeg_path
        self.mp4box_path = mp4box_path
        self.download_mode = download_mode
        self.remux_mode = remux_mode
        self.cover_format = cover_format
        self.template_folder_album = template_folder_album
        self.template_folder_compilation = template_folder_compilation
        self.template_file_single_disc = template_file_single_disc
        self.template_file_multi_disc = template_file_multi_disc
        self.template_folder_no_album = template_folder_no_album
        self.template_file_no_album = template_file_no_album
        self.template_date = template_date
        self.exclude_tags = exclude_tags
        self.cover_size = cover_size
        self.truncate = truncate
        self.no_progress = no_progress
        self._set_binaries_path_full()
        self._set_exclude_tags_list()
        self._set_truncate()

    def _set_binaries_path_full(self):
        self.nm3u8dlre_path_full = shutil.which(self.nm3u8dlre_path)
        self.ffmpeg_path_full = shutil.which(self.ffmpeg_path)
        self.mp4box_path_full = shutil.which(self.mp4box_path)
        self.mp4decrypt_path_full = shutil.which(self.mp4decrypt_path)

    def _set_exclude_tags_list(self):
        self.exclude_tags_list = (
            [i.lower() for i in self.exclude_tags.split(",")]
            if self.exclude_tags is not None
            else []
        )

    def _set_truncate(self):
        self.truncate = None if self.truncate < 4 else self.truncate

    def set_cdm(self):
        if self.wvd_path:
            self.cdm = Cdm.from_device(Device.load(self.wvd_path))
        else:
            self.cdm = Cdm.from_device(Device.loads(HARDCODED_WVD))

    def get_url_info(self, url: str) -> UrlInfo:
        url_info = UrlInfo()
        url_regex_result = re.search(
            r"/([a-z]{2})/(album|playlist|song|music-video|post)/([^/]*)(?:/([^/?]*))?(?:\?i=)?([0-9a-z]*)?",
            url,
        )
        url_info.storefront = url_regex_result.group(1)
        url_info.type = (
            "song" if url_regex_result.group(5) else url_regex_result.group(2)
        )
        url_info.id = (
            url_regex_result.group(5)
            or url_regex_result.group(4)
            or url_regex_result.group(3)
        )
        return url_info

    def get_download_queue(self, url_info: UrlInfo) -> list[DownloadQueueItem]:
        return self._get_download_queue(url_info.type, url_info.id)

    def _get_download_queue(self, url_type: str, id: str) -> list[DownloadQueueItem]:
        download_queue = []
        if url_type == "song":
            download_queue.append(DownloadQueueItem(self.apple_music_api.get_song(id)))
        elif url_type == "album":
            album = self.apple_music_api.get_album(id)
            download_queue.extend(
                DownloadQueueItem(track)
                for track in album["relationships"]["tracks"]["data"]
            )
        elif url_type == "playlist":
            download_queue.extend(
                DownloadQueueItem(track)
                for track in self.apple_music_api.get_playlist(id)["relationships"][
                    "tracks"
                ]["data"]
            )
        elif url_type == "music-video":
            download_queue.append(
                DownloadQueueItem(self.apple_music_api.get_music_video(id))
            )
        elif url_type == "post":
            download_queue.append(DownloadQueueItem(self.apple_music_api.get_post(id)))
        else:
            raise Exception(f"Invalid url type: {url_type}")
        return download_queue

    def sanitize_date(self, date: str):
        datetime_obj = ciso8601.parse_datetime(date)
        return datetime_obj.strftime(self.template_date)

    def get_decryption_key(self, pssh: str, track_id: str) -> str:
        pssh_obj = PSSH(pssh.split(",")[-1])
        cdm_session = self.cdm.open()
        challenge = base64.b64encode(
            self.cdm.get_license_challenge(cdm_session, pssh_obj)
        ).decode()
        license = self.apple_music_api.get_widevine_license(
            track_id,
            pssh,
            challenge,
        )
        self.cdm.parse_license(cdm_session, license)
        decryption_key = next(
            i for i in self.cdm.get_keys(cdm_session) if i.type == "CONTENT"
        ).key.hex()
        self.cdm.close(cdm_session)
        return decryption_key

    def download(self, path: Path, stream_url: str):
        if self.download_mode == DownloadMode.YTDLP:
            self.download_ytdlp(path, stream_url)
        elif self.download_mode == DownloadMode.NM3U8DLRE:
            self.download_nm3u8dlre(path, stream_url)

    def download_ytdlp(self, path: Path, stream_url: str):
        with YoutubeDL(
            {
                "quiet": True,
                "no_warnings": True,
                "outtmpl": str(path),
                "allow_unplayable_formats": True,
                "fixup": "never",
                "allowed_extractors": ["generic"],
                "noprogress": self.no_progress,
            }
        ) as ydl:
            ydl.download(stream_url)

    def download_nm3u8dlre(self, path: Path, stream_url: str):
        if self.no_progress:
            subprocess_additional_args = {
                "stdout": subprocess.DEVNULL,
                "stderr": subprocess.DEVNULL,
            }
        else:
            subprocess_additional_args = {}
        path.parent.mkdir(parents=True, exist_ok=True)
        subprocess.run(
            [
                self.nm3u8dlre_path_full,
                stream_url,
                "--binary-merge",
                "--no-log",
                "--log-level",
                "off",
                "--ffmpeg-binary-path",
                self.ffmpeg_path_full,
                "--save-name",
                path.stem,
                "--save-dir",
                path.parent,
                "--tmp-dir",
                path.parent,
            ],
            check=True,
            **subprocess_additional_args,
        )

    def get_sanitized_string(self, dirty_string: str, is_folder: bool) -> str:
        dirty_string = re.sub(self.ILLEGAL_CHARACTERS_REGEX, "_", dirty_string)
        if is_folder:
            dirty_string = dirty_string[: self.truncate]
            if dirty_string.endswith("."):
                dirty_string = dirty_string[:-1] + "_"
        else:
            if self.truncate is not None:
                dirty_string = dirty_string[: self.truncate - 4]
        return dirty_string.strip()

    def get_final_path(self, tags: dict, file_extension: str) -> Path:
        if tags.get("album"):
            final_path_folder = (
                self.template_folder_compilation.split("/")
                if tags.get("compilation")
                else self.template_folder_album.split("/")
            )
            final_path_file = (
                self.template_file_multi_disc.split("/")
                if tags["disc_total"] > 1
                else self.template_file_single_disc.split("/")
            )
        else:
            final_path_folder = self.template_folder_no_album.split("/")
            final_path_file = self.template_file_no_album.split("/")
        final_path_folder = [
            self.get_sanitized_string(i.format(**tags), True) for i in final_path_folder
        ]
        final_path_file = [
            self.get_sanitized_string(i.format(**tags), True)
            for i in final_path_file[:-1]
        ] + [
            self.get_sanitized_string(final_path_file[-1].format(**tags), False)
            + file_extension
        ]
        return self.output_path.joinpath(*final_path_folder).joinpath(*final_path_file)

    def get_cover_url(self, metadata: dict) -> str:
        return self._get_cover_url(metadata["attributes"]["artwork"]["url"])

    def _get_cover_url(self, cover_url_template: str) -> str:
        return re.sub(
            r"\{w\}x\{h\}([a-z]{2})\.jpg",
            f"{self.cover_size}x{self.cover_size}bb.{self.cover_format.value}",
            cover_url_template,
        )

    @staticmethod
    @functools.lru_cache()
    def get_url_response_bytes(url: str) -> bytes:
        return requests.get(url).content

    def apply_tags(
        self,
        path: Path,
        tags: dict,
        cover_url: str,
    ):
        to_apply_tags = [
            tag_name
            for tag_name in tags.keys()
            if tag_name not in self.exclude_tags_list
        ]
        mp4_tags = {}
        for tag_name in to_apply_tags:
            if tag_name in ("disc", "disc_total"):
                if mp4_tags.get("disk") is None:
                    mp4_tags["disk"] = [[0, 0]]
                if tag_name == "disc":
                    mp4_tags["disk"][0][0] = tags[tag_name]
                elif tag_name == "disc_total":
                    mp4_tags["disk"][0][1] = tags[tag_name]
            elif tag_name in ("track", "track_total"):
                if mp4_tags.get("trkn") is None:
                    mp4_tags["trkn"] = [[0, 0]]
                if tag_name == "track":
                    mp4_tags["trkn"][0][0] = tags[tag_name]
                elif tag_name == "track_total":
                    mp4_tags["trkn"][0][1] = tags[tag_name]
            elif tag_name == "compilation":
                mp4_tags["cpil"] = tags["compilation"]
            elif tag_name == "gapless":
                mp4_tags["pgap"] = tags["gapless"]
            elif (
                MP4_TAGS_MAP.get(tag_name) is not None
                and tags.get(tag_name) is not None
            ):
                mp4_tags[MP4_TAGS_MAP[tag_name]] = [tags[tag_name]]
        if "cover" not in self.exclude_tags_list:
            mp4_tags["covr"] = [
                MP4Cover(
                    self.get_url_response_bytes(cover_url),
                    imageformat=(
                        MP4Cover.FORMAT_JPEG
                        if self.cover_format == CoverFormat.JPG
                        else MP4Cover.FORMAT_PNG
                    ),
                )
            ]
        mp4 = MP4(path)
        mp4.clear()
        mp4.update(mp4_tags)
        mp4.save()

    def move_to_output_path(
        self,
        remuxed_path: Path,
        final_path: Path,
    ):
        final_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.move(remuxed_path, final_path)

    @functools.lru_cache()
    def save_cover(self, cover_path: Path, cover_url: str):
        cover_path.write_bytes(self.get_url_response_bytes(cover_url))

    def cleanup_temp_path(self):
        shutil.rmtree(self.temp_path)



from dataclasses import dataclass


@dataclass
class UrlInfo:
    storefront: str = None
    type: str = None
    id: str = None


@dataclass
class DownloadQueueItem:
    metadata: dict = None


@dataclass
class Lyrics:
    synced: str = None
    unsynced: str = None


@dataclass
class StreamInfo:
    stream_url: str = None
    pssh: str = None
    codec: str = None



from __future__ import annotations

import functools
import re
import time
from http.cookiejar import MozillaCookieJar
from pathlib import Path

import requests


class AppleMusicApi:
    APPLE_MUSIC_HOMEPAGE_URL = "https://beta.music.apple.com"
    AMP_API_URL = "https://amp-api.music.apple.com"
    WEBPLAYBACK_API_URL = (
        "https://play.itunes.apple.com/WebObjects/MZPlay.woa/wa/webPlayback"
    )
    LICENSE_API_URL = "https://play.itunes.apple.com/WebObjects/MZPlay.woa/wa/acquireWebPlaybackLicense"
    WAIT_TIME = 2

    def __init__(
        self,
        cookies_path: Path | None = Path("./cookies.txt"),
        storefront: None | str = None,
        language: str = "en-US",
    ):
        self.cookies_path = cookies_path
        self.storefront = storefront
        self.language = language
        self._set_session()

    def _set_session(self):
        self.session = requests.Session()
        if self.cookies_path:
            cookies = MozillaCookieJar(self.cookies_path)
            cookies.load(ignore_discard=True, ignore_expires=True)
            self.session.cookies.update(cookies)
            self.storefront = self.session.cookies.get_dict()["itua"]
        self.session.headers.update(
            {
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:95.0) Gecko/20100101 Firefox/95.0",
                "Accept": "application/json",
                "Accept-Language": "en-US,en;q=0.5",
                "Accept-Encoding": "gzip, deflate, br",
                "content-type": "application/json",
                "Media-User-Token": self.session.cookies.get_dict().get(
                    "media-user-token", ""
                ),
                "x-apple-renewal": "true",
                "DNT": "1",
                "Connection": "keep-alive",
                "Sec-Fetch-Dest": "empty",
                "Sec-Fetch-Mode": "cors",
                "Sec-Fetch-Site": "same-site",
                "origin": self.APPLE_MUSIC_HOMEPAGE_URL,
            }
        )
        home_page = self.session.get(self.APPLE_MUSIC_HOMEPAGE_URL).text
        index_js_uri = re.search(
            r"/(assets/index-legacy-[^/]+\.js)",
            home_page,
        ).group(1)
        index_js_page = self.session.get(
            f"{self.APPLE_MUSIC_HOMEPAGE_URL}/{index_js_uri}"
        ).text
        token = re.search('(?=eyJh)(.*?)(?=")', index_js_page).group(1)
        self.session.headers.update({"authorization": f"Bearer {token}"})
        self.session.params = {"l": self.language}

    @staticmethod
    def _raise_response_exception(response: requests.Response):
        raise Exception(
            f"Request failed with status code {response.status_code}: {response.text}"
        )

    def _check_amp_api_response(self, response: requests.Response):
        try:
            response.raise_for_status()
            response_dict = response.json()
            assert response_dict.get("data")
        except (
            requests.HTTPError,
            requests.exceptions.JSONDecodeError,
            AssertionError,
        ):
            self._raise_response_exception(response)

    def get_song(
        self,
        song_id: str,
        extend: str = "extendedAssetUrls",
        include: str = "lyrics,albums",
    ) -> dict:
        response = self.session.get(
            f"{self.AMP_API_URL}/v1/catalog/{self.storefront}/songs/{song_id}",
            params={
                "include": include,
                "extend": extend,
            },
        )
        self._check_amp_api_response(response)
        return response.json()["data"][0]

    def get_music_video(
        self,
        music_video_id: str,
        include: str = "albums",
    ) -> dict:
        response = self.session.get(
            f"{self.AMP_API_URL}/v1/catalog/{self.storefront}/music-videos/{music_video_id}",
            params={
                "include": include,
            },
        )
        self._check_amp_api_response(response)
        return response.json()["data"][0]

    def get_post(
        self,
        post_id: str,
    ) -> dict:
        response = self.session.get(
            f"{self.AMP_API_URL}/v1/catalog/{self.storefront}/uploaded-videos/{post_id}"
        )
        self._check_amp_api_response(response)
        return response.json()["data"][0]

    @functools.lru_cache()
    def get_album(
        self,
        album_id: str,
        extend: str = "extendedAssetUrls",
    ) -> dict:
        response = self.session.get(
            f"{self.AMP_API_URL}/v1/catalog/{self.storefront}/albums/{album_id}",
            params={
                "extend": extend,
            },
        )
        self._check_amp_api_response(response)
        return response.json()["data"][0]

    def get_playlist(
        self,
        playlist_id: str,
        is_library: bool = False,
        limit_tracks: int = 300,
        extend: str = "extendedAssetUrls",
        full_playlist: bool = True,
    ) -> dict:
        response = self.session.get(
            f"{self.AMP_API_URL}/v1/{'me' if is_library else 'catalog'}/{self.storefront}/playlists/{playlist_id}",
            params={
                "extend": extend,
                "limit[tracks]": limit_tracks,
            },
        )
        self._check_amp_api_response(response)
        playlist = response.json()["data"][0]
        if full_playlist:
            playlist = self._extend_playlists_tracks(playlist, limit_tracks)
        return playlist

    def _extend_playlists_tracks(
        self,
        playlist: dict,
        limit_tracks: int,
    ) -> dict:
        playlist_next_uri = playlist["relationships"]["tracks"].get("next")
        while playlist_next_uri:
            playlist_next = self._get_playlist_next(playlist_next_uri, limit_tracks)
            playlist["relationships"]["tracks"]["data"].extend(playlist_next["data"])
            playlist_next_uri = playlist_next.get("next")
            time.sleep(self.WAIT_TIME)
        return playlist

    def _get_playlist_next(self, playlist_next_uri: str, limit_tracks: int) -> dict:
        response = self.session.get(
            self.AMP_API_URL + playlist_next_uri,
            params={
                "limit[tracks]": limit_tracks,
            },
        )
        self._check_amp_api_response(response)
        return response.json()

    def get_webplayback(
        self,
        track_id: str,
    ) -> dict:
        response = self.session.post(
            self.WEBPLAYBACK_API_URL,
            json={
                "salableAdamId": track_id,
                "language": self.language,
            },
        )
        try:
            response.raise_for_status()
            response_dict = response.json()
            webplayback = response_dict.get("songList")
            assert webplayback
        except (
            requests.HTTPError,
            requests.exceptions.JSONDecodeError,
            AssertionError,
        ):
            self._raise_response_exception(response)
        return webplayback[0]

    def get_widevine_license(
        self,
        track_id: str,
        track_uri: str,
        challenge: str,
    ) -> str:
        response = self.session.post(
            self.LICENSE_API_URL,
            json={
                "challenge": challenge,
                "key-system": "com.widevine.alpha",
                "uri": track_uri,
                "adamId": track_id,
                "isLibrary": False,
                "user-initiated": True,
            },
        )
        try:
            response.raise_for_status()
            response_dict = response.json()
            widevine_license = response_dict.get("license")
            assert widevine_license
        except (
            requests.HTTPError,
            requests.exceptions.JSONDecodeError,
            AssertionError,
        ):
            self._raise_response_exception(response)
        return widevine_license



from __future__ import annotations

import inspect
import json
import logging
from enum import Enum
from pathlib import Path

import click

from . import __version__
from .apple_music_api import AppleMusicApi
from .constants import *
from .downloader import Downloader
from .downloader_music_video import DownloaderMusicVideo
from .downloader_post import DownloaderPost
from .downloader_song import DownloaderSong
from .downloader_song_legacy import DownloaderSongLegacy
from .enums import CoverFormat, DownloadMode, MusicVideoCodec, PostQuality, RemuxMode
from .itunes_api import ItunesApi

apple_music_api_sig = inspect.signature(AppleMusicApi.__init__)
downloader_sig = inspect.signature(Downloader.__init__)
downloader_song_sig = inspect.signature(DownloaderSong.__init__)
downloader_music_video_sig = inspect.signature(DownloaderMusicVideo.__init__)
downloader_post_sig = inspect.signature(DownloaderPost.__init__)


def get_param_string(param: click.Parameter) -> str:
    if isinstance(param.default, Enum):
        return param.default.value
    elif isinstance(param.default, Path):
        return str(param.default)
    else:
        return param.default


def write_default_config_file(ctx: click.Context) -> None:
    ctx.params["config_path"].parent.mkdir(parents=True, exist_ok=True)
    config_file = {
        param.name: get_param_string(param)
        for param in ctx.command.params
        if param.name not in EXCLUDED_CONFIG_FILE_PARAMS
    }
    ctx.params["config_path"].write_text(json.dumps(config_file, indent=4))


def load_config_file(
    ctx: click.Context,
    param: click.Parameter,
    no_config_file: bool,
) -> click.Context:
    if no_config_file:
        return ctx
    if not ctx.params["config_path"].exists():
        write_default_config_file(ctx)
    config_file = dict(json.loads(ctx.params["config_path"].read_text()))
    for param in ctx.command.params:
        if (
            config_file.get(param.name) is not None
            and not ctx.get_parameter_source(param.name)
            == click.core.ParameterSource.COMMANDLINE
        ):
            ctx.params[param.name] = param.type_cast_value(ctx, config_file[param.name])
    return ctx


@click.command()
@click.help_option("-h", "--help")
@click.version_option(__version__, "-v", "--version")
# CLI specific options
@click.argument(
    "urls",
    nargs=-1,
    type=str,
    required=True,
)
@click.option(
    "--disable-music-video-skip",
    is_flag=True,
    help="Don't skip downloading music videos in albums/playlists.",
)
@click.option(
    "--save-cover",
    "-s",
    is_flag=True,
    help="Save cover as a separate file.",
)
@click.option(
    "--overwrite",
    is_flag=True,
    help="Overwrite existing files.",
)
@click.option(
    "--read-urls-as-txt",
    "-r",
    is_flag=True,
    help="Interpret URLs as paths to text files containing URLs.",
)
@click.option(
    "--synced-lyrics-only",
    is_flag=True,
    help="Download only the synced lyrics.",
)
@click.option(
    "--no-synced-lyrics",
    is_flag=True,
    help="Don't download the synced lyrics.",
)
@click.option(
    "--config-path",
    type=Path,
    default=Path.home() / ".gamdl" / "config.json",
    help="Path to config file.",
)
@click.option(
    "--log-level",
    type=str,
    default="INFO",
    help="Log level.",
)
@click.option(
    "--print-exceptions",
    is_flag=True,
    help="Print exceptions.",
)
# API specific options
@click.option(
    "--cookies-path",
    "-c",
    type=Path,
    default=apple_music_api_sig.parameters["cookies_path"].default,
    help="Path to .txt cookies file.",
)
# Downloader specific options
@click.option(
    "--output-path",
    "-o",
    type=Path,
    default=downloader_sig.parameters["output_path"].default,
    help="Path to output directory.",
)
@click.option(
    "--temp-path",
    type=Path,
    default=downloader_sig.parameters["temp_path"].default,
    help="Path to temporary directory.",
)
@click.option(
    "--wvd-path",
    type=Path,
    default=downloader_sig.parameters["wvd_path"].default,
    help="Path to .wvd file.",
)
@click.option(
    "--nm3u8dlre-path",
    type=str,
    default=downloader_sig.parameters["nm3u8dlre_path"].default,
    help="Path to N_m3u8DL-RE binary.",
)
@click.option(
    "--mp4decrypt-path",
    type=str,
    default=downloader_sig.parameters["mp4decrypt_path"].default,
    help="Path to mp4decrypt binary.",
)
@click.option(
    "--ffmpeg-path",
    type=str,
    default=downloader_sig.parameters["ffmpeg_path"].default,
    help="Path to FFmpeg binary.",
)
@click.option(
    "--mp4box-path",
    type=str,
    default=downloader_sig.parameters["mp4box_path"].default,
    help="Path to MP4Box binary.",
)
@click.option(
    "--download-mode",
    type=DownloadMode,
    default=downloader_sig.parameters["download_mode"].default,
    help="Download mode.",
)
@click.option(
    "--remux-mode",
    type=RemuxMode,
    default=downloader_sig.parameters["remux_mode"].default,
    help="Remux mode.",
)
@click.option(
    "--cover-format",
    type=CoverFormat,
    default=downloader_sig.parameters["cover_format"].default,
    help="Cover format.",
)
@click.option(
    "--template-folder-album",
    type=str,
    default=downloader_sig.parameters["template_folder_album"].default,
    help="Template folder for tracks that are part of an album.",
)
@click.option(
    "--template-folder-compilation",
    type=str,
    default=downloader_sig.parameters["template_folder_compilation"].default,
    help="Template folder for tracks that are part of a compilation album.",
)
@click.option(
    "--template-file-single-disc",
    type=str,
    default=downloader_sig.parameters["template_file_single_disc"].default,
    help="Template file for the tracks that are part of a single-disc album.",
)
@click.option(
    "--template-file-multi-disc",
    type=str,
    default=downloader_sig.parameters["template_file_multi_disc"].default,
    help="Template file for the tracks that are part of a multi-disc album.",
)
@click.option(
    "--template-folder-no-album",
    type=str,
    default=downloader_sig.parameters["template_folder_no_album"].default,
    help="Template folder for the tracks that are not part of an album.",
)
@click.option(
    "--template-file-no-album",
    type=str,
    default=downloader_sig.parameters["template_file_no_album"].default,
    help="Template file for the tracks that are not part of an album.",
)
@click.option(
    "--template-date",
    type=str,
    default=downloader_sig.parameters["template_date"].default,
    help="Date tag template.",
)
@click.option(
    "--exclude-tags",
    type=str,
    default=downloader_sig.parameters["exclude_tags"].default,
    help="Comma-separated tags to exclude.",
)
@click.option(
    "--cover-size",
    type=int,
    default=downloader_sig.parameters["cover_size"].default,
    help="Cover size.",
)
@click.option(
    "--truncate",
    type=int,
    default=downloader_sig.parameters["truncate"].default,
    help="Maximum length of the file/folder names.",
)
# DownloaderSong specific options
@click.option(
    "--codec-song",
    type=SongCodec,
    default=downloader_song_sig.parameters["codec"].default,
    help="Song codec.",
)
@click.option(
    "--synced-lyrics-format",
    type=SyncedLyricsFormat,
    default=downloader_song_sig.parameters["synced_lyrics_format"].default,
    help="Synced lyrics format.",
)
# DownloaderMusicVideo specific options
@click.option(
    "--codec-music-video",
    type=MusicVideoCodec,
    default=downloader_music_video_sig.parameters["codec"].default,
    help="Music video codec.",
)
# DownloaderPost specific options
@click.option(
    "--quality-post",
    type=PostQuality,
    default=downloader_post_sig.parameters["quality"].default,
    help="Post video quality.",
)
# This option should always be last
@click.option(
    "--no-config-file",
    "-n",
    is_flag=True,
    callback=load_config_file,
    help="Do not use a config file.",
)
def main(
    urls: list[str],
    disable_music_video_skip: bool,
    save_cover: bool,
    overwrite: bool,
    read_urls_as_txt: bool,
    synced_lyrics_only: bool,
    no_synced_lyrics: bool,
    config_path: Path,
    log_level: str,
    print_exceptions: bool,
    cookies_path: Path,
    output_path: Path,
    temp_path: Path,
    wvd_path: Path,
    nm3u8dlre_path: str,
    mp4decrypt_path: str,
    ffmpeg_path: str,
    mp4box_path: str,
    download_mode: DownloadMode,
    remux_mode: RemuxMode,
    cover_format: CoverFormat,
    template_folder_album: str,
    template_folder_compilation: str,
    template_file_single_disc: str,
    template_file_multi_disc: str,
    template_folder_no_album: str,
    template_file_no_album: str,
    template_date: str,
    exclude_tags: str,
    cover_size: int,
    truncate: int,
    codec_song: SongCodec,
    synced_lyrics_format: SyncedLyricsFormat,
    codec_music_video: MusicVideoCodec,
    quality_post: PostQuality,
    no_config_file: bool,
):
    logging.basicConfig(
        format="[%(levelname)-8s %(asctime)s] %(message)s",
        datefmt="%H:%M:%S",
    )
    logger = logging.getLogger(__name__)
    logger.setLevel(log_level)
    logger.debug("Starting downloader")
    apple_music_api = AppleMusicApi(cookies_path)
    itunes_api = ItunesApi(
        apple_music_api.storefront,
        apple_music_api.language,
    )
    downloader = Downloader(
        apple_music_api,
        itunes_api,
        output_path,
        temp_path,
        wvd_path,
        nm3u8dlre_path,
        mp4decrypt_path,
        ffmpeg_path,
        mp4box_path,
        download_mode,
        remux_mode,
        cover_format,
        template_folder_album,
        template_folder_compilation,
        template_file_single_disc,
        template_file_multi_disc,
        template_folder_no_album,
        template_file_no_album,
        template_date,
        exclude_tags,
        cover_size,
        truncate,
    )
    downloader_song = DownloaderSong(
        downloader,
        codec_song,
        synced_lyrics_format,
    )
    downloader_song_legacy = DownloaderSongLegacy(
        downloader,
        codec_song,
    )
    downloader_music_video = DownloaderMusicVideo(
        downloader,
        codec_music_video,
    )
    downloader_post = DownloaderPost(
        downloader,
        quality_post,
    )
    if not synced_lyrics_only:
        if wvd_path and not wvd_path.exists():
            logger.critical(X_NOT_FOUND_STRING.format(".wvd file", wvd_path))
            return
        logger.debug("Setting up CDM")
        downloader.set_cdm()
        if not downloader.ffmpeg_path_full and (
            remux_mode == RemuxMode.FFMPEG or download_mode == DownloadMode.NM3U8DLRE
        ):
            logger.critical(X_NOT_FOUND_STRING.format("ffmpeg", ffmpeg_path))
            return
        if not downloader.mp4box_path_full and remux_mode == RemuxMode.MP4BOX:
            logger.critical(X_NOT_FOUND_STRING.format("MP4Box", mp4box_path))
            return
        if (
            not downloader.mp4decrypt_path_full
            and codec_song
            not in (
                SongCodec.AAC_LEGACY,
                SongCodec.AAC_HE_LEGACY,
            )
            or (remux_mode == RemuxMode.MP4BOX and not downloader.mp4decrypt_path_full)
        ):
            logger.critical(X_NOT_FOUND_STRING.format("mp4decrypt", mp4decrypt_path))
            return
        if (
            download_mode == DownloadMode.NM3U8DLRE
            and not downloader.nm3u8dlre_path_full
        ):
            logger.critical(X_NOT_FOUND_STRING.format("N_m3u8DL-RE", nm3u8dlre_path))
            return
        if not downloader.mp4decrypt_path_full:
            logger.warn(
                X_NOT_FOUND_STRING.format("mp4decrypt", mp4decrypt_path)
                + ", music videos will not be downloaded"
            )
            skip_mv = True
        else:
            skip_mv = False
        if codec_song not in LEGACY_CODECS:
            logger.warn(
                "You have chosen a non-legacy codec. Support for non-legacy codecs are not guaranteed, "
                "as most of the songs cannot be downloaded when using non-legacy codecs."
            )
    error_count = 0
    if read_urls_as_txt:
        urls = [url.strip() for url in Path(urls[0]).read_text().splitlines()]
    for url_index, url in enumerate(urls, start=1):
        url_progress = f"URL {url_index}/{len(urls)}"
        try:
            url_info = downloader.get_url_info(url)
            download_queue = downloader.get_download_queue(url_info)
        except Exception as e:
            error_count += 1
            logger.error(
                f'({url_progress}) Failed to check "{url}"',
                exc_info=print_exceptions,
            )
            continue
        for queue_index, queue_item in enumerate(download_queue, start=1):
            queue_progress = f"Track {queue_index}/{len(download_queue)} from URL {url_index}/{len(urls)}"
            track = queue_item.metadata
            try:
                logger.info(
                    f'({queue_progress}) Downloading "{track["attributes"]["name"]}"'
                )
                if not track["attributes"].get("playParams"):
                    logger.warning(
                        f"({queue_progress}) Track is not streamable, skipping"
                    )
                    continue
                if (
                    (synced_lyrics_only and track["type"] != "songs")
                    or (track["type"] == "music-videos" and skip_mv)
                    or (
                        track["type"] == "music-videos"
                        and url_info.type == "album"
                        and not disable_music_video_skip
                    )
                ):
                    logger.warning(
                        f"({queue_progress}) Track is not downloadable with current configuration, skipping"
                    )
                elif track["type"] == "songs":
                    logger.debug("Getting lyrics")
                    lyrics = downloader_song.get_lyrics(track)
                    logger.debug("Getting webplayback")
                    webplayback = apple_music_api.get_webplayback(track["id"])
                    tags = downloader_song.get_tags(webplayback, lyrics.unsynced)
                    final_path = downloader.get_final_path(tags, ".m4a")
                    lyrics_synced_path = downloader_song.get_lyrics_synced_path(
                        final_path
                    )
                    cover_path = downloader_song.get_cover_path(final_path)
                    cover_url = downloader.get_cover_url(track)
                    if synced_lyrics_only:
                        pass
                    elif final_path.exists() and not overwrite:
                        logger.warning(
                            f'({queue_progress}) Song already exists at "{final_path}", skipping'
                        )
                    else:
                        if codec_song in LEGACY_CODECS:
                            logger.debug("Getting stream info")
                            stream_info = downloader_song_legacy.get_stream_info(
                                webplayback
                            )
                            logger.debug("Getting decryption key")
                            decryption_key = downloader_song_legacy.get_decryption_key(
                                stream_info.pssh, track["id"]
                            )
                        else:
                            stream_info = downloader_song.get_stream_info(track)
                            if not stream_info.stream_url or not stream_info.codec:
                                logger.warning(
                                    f"({queue_progress}) Song is not downloadable or is not"
                                    " available in the chosen codec, skipping"
                                )
                                continue
                            logger.debug("Getting decryption key")
                            decryption_key = downloader.get_decryption_key(
                                stream_info.pssh, track["id"]
                            )
                        encrypted_path = downloader_song.get_encrypted_path(track["id"])
                        decrypted_path = downloader_song.get_decrypted_path(track["id"])
                        remuxed_path = downloader_song.get_remuxed_path(track["id"])
                        logger.debug(f"Downloading to {encrypted_path}")
                        downloader.download(encrypted_path, stream_info.stream_url)
                        if codec_song in LEGACY_CODECS:
                            logger.debug(f"Remuxing/Decrypting to {remuxed_path}")
                            downloader_song_legacy.remux(
                                encrypted_path,
                                decrypted_path,
                                remuxed_path,
                                decryption_key,
                            )
                        else:
                            logger.debug(f"Decrypting to {decrypted_path}")
                            downloader_song.decrypt(
                                encrypted_path, decrypted_path, decryption_key
                            )
                            logger.debug(f"Remuxing to {final_path}")
                            downloader_song.remux(
                                decrypted_path,
                                remuxed_path,
                                stream_info.codec,
                            )
                        logger.debug("Applying tags")
                        downloader.apply_tags(remuxed_path, tags, cover_url)
                        logger.debug(f"Moving to {final_path}")
                        downloader.move_to_output_path(remuxed_path, final_path)
                    if no_synced_lyrics or not lyrics.synced:
                        pass
                    elif lyrics_synced_path.exists() and not overwrite:
                        logger.debug(
                            f'Synced lyrics already exists at "{lyrics_synced_path}", skipping'
                        )
                    else:
                        logger.debug(f'Saving synced lyrics to "{lyrics_synced_path}"')
                        downloader_song.save_lyrics_synced(
                            lyrics_synced_path, lyrics.synced
                        )
                    if synced_lyrics_only or not save_cover:
                        pass
                    elif cover_path.exists() and not overwrite:
                        logger.debug(
                            f'Cover already exists at "{cover_path}", skipping'
                        )
                    else:
                        logger.debug(f'Saving cover to "{cover_path}"')
                        downloader.save_cover(cover_path, cover_url)
                elif track["type"] == "music-videos":
                    music_video_id_alt = downloader_music_video.get_music_video_id_alt(
                        track
                    )
                    logger.debug("Getting iTunes page")
                    itunes_page = itunes_api.get_itunes_page(
                        "music-video", music_video_id_alt
                    )
                    stream_url_master = downloader_music_video.get_stream_url_master(
                        itunes_page
                    )
                    logger.debug("Getting M3U8 data")
                    m3u8_master_data = downloader_music_video.get_m3u8_master_data(
                        stream_url_master
                    )
                    tags = downloader_music_video.get_tags(
                        itunes_page,
                        m3u8_master_data,
                        track,
                    )
                    final_path = downloader.get_final_path(tags, ".m4v")
                    cover_path = downloader_music_video.get_cover_path(final_path)
                    cover_url = downloader.get_cover_url(track)
                    if final_path.exists() and not overwrite:
                        logger.warning(
                            f'({queue_progress}) Music video already exists at "{final_path}", skipping'
                        )
                    else:
                        logger.debug("Getting stream info")
                        stream_info_video, stream_info_audio = (
                            downloader_music_video.get_stream_info_video(
                                m3u8_master_data
                            ),
                            downloader_music_video.get_stream_info_audio(
                                m3u8_master_data
                            ),
                        )
                        decryption_key_video = downloader.get_decryption_key(
                            stream_info_video.pssh, track["id"]
                        )
                        decryption_key_audio = downloader.get_decryption_key(
                            stream_info_audio.pssh, track["id"]
                        )
                        encrypted_path_video = (
                            downloader_music_video.get_encrypted_path_video(track["id"])
                        )
                        encrypted_path_audio = (
                            downloader_music_video.get_encrypted_path_audio(track["id"])
                        )
                        decrypted_path_video = (
                            downloader_music_video.get_decrypted_path_video(track["id"])
                        )
                        decrypted_path_audio = (
                            downloader_music_video.get_decrypted_path_audio(track["id"])
                        )
                        remuxed_path = downloader_music_video.get_remuxed_path(
                            track["id"]
                        )
                        logger.debug(f"Downloading video to {encrypted_path_video}")
                        downloader.download(
                            encrypted_path_video, stream_info_video.stream_url
                        )
                        logger.debug(f"Downloading audio to {encrypted_path_audio}")
                        downloader.download(
                            encrypted_path_audio, stream_info_audio.stream_url
                        )
                        logger.debug(f"Decrypting video to {decrypted_path_video}")
                        downloader_music_video.decrypt(
                            encrypted_path_video,
                            decryption_key_video,
                            decrypted_path_video,
                        )
                        logger.debug(f"Decrypting audio to {decrypted_path_audio}")
                        downloader_music_video.decrypt(
                            encrypted_path_audio,
                            decryption_key_audio,
                            decrypted_path_audio,
                        )
                        logger.debug(f"Remuxing to {remuxed_path}")
                        downloader_music_video.remux(
                            decrypted_path_video,
                            decrypted_path_audio,
                            remuxed_path,
                            stream_info_video.codec,
                            stream_info_audio.codec,
                        )
                        logger.debug("Applying tags")
                        downloader.apply_tags(remuxed_path, tags, cover_url)
                        logger.debug(f"Moving to {final_path}")
                        downloader.move_to_output_path(remuxed_path, final_path)
                    if not save_cover:
                        pass
                    elif cover_path.exists() and not overwrite:
                        logger.debug(
                            f'Cover already exists at "{cover_path}", skipping'
                        )
                    else:
                        logger.debug(f'Saving cover to "{cover_path}"')
                        downloader.save_cover(cover_path, cover_url)
                elif track["type"] == "uploaded-videos":
                    stream_url = downloader_post.get_stream_url(track)
                    tags = downloader_post.get_tags(track)
                    temp_path = downloader_post.get_temp_path(track["id"])
                    final_path = downloader.get_final_path(tags, ".m4v")
                    cover_path = downloader_music_video.get_cover_path(final_path)
                    cover_url = downloader.get_cover_url(track)
                    if final_path.exists() and not overwrite:
                        logger.warning(
                            f'({queue_progress}) Post video already exists at "{final_path}", skipping'
                        )
                    else:
                        logger.debug(f"Downloading to {final_path}")
                        downloader.download_ytdlp(temp_path, stream_url)
                        logger.debug("Applying tags")
                        downloader.apply_tags(temp_path, tags, cover_url)
                        logger.debug(f"Moving to {final_path}")
                        downloader.move_to_output_path(temp_path, final_path)
                    if not save_cover:
                        pass
                    elif cover_path.exists() and not overwrite:
                        logger.debug(
                            f'Cover already exists at "{cover_path}", skipping'
                        )
                    else:
                        logger.debug(f'Saving cover to "{cover_path}"')
                        downloader.save_cover(cover_path, cover_url)
            except Exception as e:
                error_count += 1
                logger.error(
                    f'({queue_progress}) Failed to download "{track["attributes"]["name"]}"',
                    exc_info=print_exceptions,
                )
            finally:
                if temp_path.exists():
                    logger.debug(f'Cleaning up "{temp_path}"')
                    downloader.cleanup_temp_path()
    logger.info(f"Done ({error_count} error(s))")



import re
import subprocess
import urllib.parse
from pathlib import Path

import click
import m3u8
from tabulate import tabulate

from .constants import MUSIC_VIDEO_CODEC_MAP
from .downloader import Downloader
from .enums import MusicVideoCodec, RemuxMode
from .models import StreamInfo


class DownloaderMusicVideo:
    MP4_FORMAT_CODECS = ["hvc1", "ec-3"]

    def __init__(
        self,
        downloader: Downloader,
        codec: MusicVideoCodec = MusicVideoCodec.H264_BEST,
    ):
        self.downloader = downloader
        self.codec = codec

    def get_stream_url_master(self, itunes_page: dict) -> str:
        return itunes_page["offers"][0]["assets"][0]["hlsUrl"]

    def get_m3u8_master_data(self, stream_url_master: str) -> dict:
        url_parts = urllib.parse.urlparse(stream_url_master)
        query = urllib.parse.parse_qs(url_parts.query, keep_blank_values=True)
        query.update({"aec": "HD", "dsid": "1"})
        stream_url_master_new = url_parts._replace(
            query=urllib.parse.urlencode(query, doseq=True)
        ).geturl()
        return m3u8.load(stream_url_master_new).data

    def get_playlist_video(
        self,
        playlists: list[dict],
    ) -> dict:
        playlists_filtered = [
            playlist
            for playlist in playlists
            if playlist["stream_info"]["codecs"].startswith(
                MUSIC_VIDEO_CODEC_MAP[self.codec]
            )
        ]
        if not playlists_filtered:
            playlists_filtered = [
                playlist
                for playlist in playlists
                if playlist["stream_info"]["codecs"].startswith(
                    MUSIC_VIDEO_CODEC_MAP[MusicVideoCodec.H264_BEST]
                )
            ]
        playlists_filtered.sort(key=lambda x: x["stream_info"]["bandwidth"])
        return playlists_filtered[-1]

    def get_playlist_video_from_user(
        self,
        playlists: list[dict],
    ) -> dict:
        table = [
            [
                i,
                playlist["stream_info"]["codecs"],
                playlist["stream_info"]["resolution"],
                playlist["stream_info"]["bandwidth"],
            ]
            for i, playlist in enumerate(playlists, 1)
        ]
        print(tabulate(table))
        try:
            choice = (
                click.prompt("Choose a video codec", type=click.IntRange(1, len(table)))
                - 1
            )
        except click.exceptions.Abort:
            raise KeyboardInterrupt()
        return playlists[choice]

    def get_playlist_audio(
        self,
        playlists: list[dict],
    ) -> dict:
        stream_url = next(
            (
                playlist
                for playlist in playlists
                if playlist["group_id"] == "audio-stereo-256"
            ),
            None,
        )
        return stream_url

    def get_playlist_audio_from_user(
        self,
        playlists: list[dict],
    ) -> dict:
        table = [
            [
                i,
                playlist["group_id"],
            ]
            for i, playlist in enumerate(playlists, 1)
        ]
        print(tabulate(table))
        try:
            choice = (
                click.prompt(
                    "Choose an audio codec", type=click.IntRange(1, len(table))
                )
                - 1
            )
        except click.exceptions.Abort:
            raise KeyboardInterrupt()
        return playlists[choice]

    def get_pssh(self, m3u8_data: dict):
        return next(
            (
                key
                for key in m3u8_data["keys"]
                if key["keyformat"] == "urn:uuid:edef8ba9-79d6-4ace-a3c8-27dcd51d21ed"
            ),
            None,
        )["uri"]

    def get_stream_info_video(self, m3u8_master_data: dict) -> StreamInfo:
        stream_info = StreamInfo()
        if self.codec != MusicVideoCodec.ASK:
            playlist = self.get_playlist_video(m3u8_master_data["playlists"])
        else:
            playlist = self.get_playlist_video_from_user(m3u8_master_data["playlists"])
        stream_info.stream_url = playlist["uri"]
        stream_info.codec = playlist["stream_info"]["codecs"]
        m3u8_data = m3u8.load(stream_info.stream_url).data
        stream_info.pssh = self.get_pssh(m3u8_data)
        return stream_info

    def get_stream_info_audio(self, m3u8_master_data: dict) -> StreamInfo:
        stream_info = StreamInfo()
        if self.codec != MusicVideoCodec.ASK:
            playlist = self.get_playlist_audio(m3u8_master_data["media"])
        else:
            playlist = self.get_playlist_audio_from_user(m3u8_master_data["media"])
        stream_info.stream_url = playlist["uri"]
        stream_info.codec = re.search(r"_([^_]+)\.m3u8", stream_info.stream_url).group(
            1
        )
        m3u8_data = m3u8.load(stream_info.stream_url).data
        stream_info.pssh = self.get_pssh(m3u8_data)
        return stream_info

    def get_music_video_id_alt(self, metadata: dict) -> str:
        return metadata["attributes"]["url"].split("/")[-1].split("?")[0]

    def get_tags(
        self,
        itunes_page: dict,
        m3u8_master_data: dict,
        metadata: dict,
    ):
        tags = {
            "artist": metadata["attributes"]["artistName"],
            "artist_id": int(itunes_page["artistId"]),
            "copyright": itunes_page["copyright"],
            "date": next(
                (
                    session_data
                    for session_data in m3u8_master_data["session_data"]
                    if session_data["data_id"] == "com.apple.hls.release-date"
                ),
                None,
            )["value"],
            "genre": metadata["attributes"]["genreNames"][0],
            "genre_id": int(itunes_page["genres"][0]["genreId"]),
            "media_type": 6,
            "title": metadata["attributes"]["name"],
            "title_id": int(metadata["id"]),
        }
        if metadata["attributes"].get("contentRating") == "clean":
            tags["rating"] = 2
        elif metadata["attributes"].get("contentRating") == "explicit":
            tags["rating"] = 1
        else:
            tags["rating"] = 0
        if itunes_page.get("collectionId"):
            metadata_itunes = self.downloader.itunes_api.get_resource(itunes_page["id"])
            album = self.downloader.apple_music_api.get_album(
                itunes_page["collectionId"]
            )
            tags["album"] = album["attributes"]["name"]
            tags["album_artist"] = album["attributes"]["artistName"]
            tags["album_id"] = int(itunes_page["collectionId"])
            tags["disc"] = metadata_itunes[0]["discNumber"]
            tags["disc_total"] = metadata_itunes[0]["discCount"]
            tags["compilation"] = album["attributes"]["isCompilation"]
            tags["track"] = metadata_itunes[0]["trackNumber"]
            tags["track_total"] = metadata_itunes[0]["trackCount"]
        return tags

    def get_encrypted_path_video(self, track_id: str) -> str:
        return self.downloader.temp_path / f"encrypted_{track_id}.mp4"

    def get_encrypted_path_audio(self, track_id: str) -> str:
        return self.downloader.temp_path / f"encrypted_{track_id}.m4a"

    def get_decrypted_path_video(self, track_id: str) -> str:
        return self.downloader.temp_path / f"decrypted_{track_id}.mp4"

    def get_decrypted_path_audio(self, track_id: str) -> str:
        return self.downloader.temp_path / f"decrypted_{track_id}.m4a"

    def get_remuxed_path(self, track_id: str) -> str:
        return self.downloader.temp_path / f"remuxed_{track_id}.m4v"

    def decrypt(self, encrypted_path: Path, decryption_key: str, decrypted_path: Path):
        subprocess.run(
            [
                self.downloader.mp4decrypt_path_full,
                encrypted_path,
                "--key",
                f"1:{decryption_key}",
                decrypted_path,
            ],
            check=True,
        )

    def remux_mp4box(
        self,
        decrypted_path_audio: Path,
        decrypted_path_video: Path,
        fixed_path: Path,
    ) -> None:
        subprocess.run(
            [
                self.downloader.mp4box_path_full,
                "-quiet",
                "-add",
                decrypted_path_audio,
                "-add",
                decrypted_path_video,
                "-itags",
                "artist=placeholder",
                "-new",
                fixed_path,
            ],
            check=True,
        )

    def remux_ffmpeg(
        self,
        decrypted_path_video: Path,
        decrypte_path_audio: Path,
        fixed_path: Path,
        codec_video: str,
        codec_audio: str,
    ):
        use_mp4_flag = any(
            codec_video.startswith(codec) for codec in self.MP4_FORMAT_CODECS
        ) or any(codec_audio.startswith(codec) for codec in self.MP4_FORMAT_CODECS)
        subprocess.run(
            [
                self.downloader.ffmpeg_path_full,
                "-loglevel",
                "error",
                "-y",
                "-i",
                decrypted_path_video,
                "-i",
                decrypte_path_audio,
                "-movflags",
                "+faststart",
                "-f",
                "mp4" if use_mp4_flag else "ipod",
                "-c",
                "copy",
                "-c:s",
                "mov_text",
                fixed_path,
            ],
            check=True,
        )

    def remux(
        self,
        decrypted_path_video: Path,
        decrypted_path_audio: Path,
        remuxed_path: Path,
        codec_video: str,
        codec_audio: str,
    ):
        if self.downloader.remux_mode == RemuxMode.MP4BOX:
            self.remux_mp4box(
                decrypted_path_audio,
                decrypted_path_video,
                remuxed_path,
            )
        elif self.downloader.remux_mode == RemuxMode.FFMPEG:
            self.remux_ffmpeg(
                decrypted_path_video,
                decrypted_path_audio,
                remuxed_path,
                codec_video,
                codec_audio,
            )

    def get_cover_path(self, final_path: Path) -> Path:
        return final_path.with_suffix(f".{self.downloader.cover_format.value}")



from __future__ import annotations

import functools

import requests

from .apple_music_api import AppleMusicApi
from .constants import STOREFRONT_IDS


class ItunesApi:
    ITUNES_LOOKUP_API_URL = "https://itunes.apple.com/lookup"
    ITUNES_PAGE_API_URL = "https://music.apple.com"

    def __init__(
        self,
        storefront: str = "us",
        language: str = "en-US",
    ):
        self.storefront = storefront
        self.language = language
        self._setup_session()

    def _setup_session(self):
        try:
            self.storefront_id = STOREFRONT_IDS[self.storefront.upper()]
        except KeyError:
            raise Exception(f"No storefront id for {self.storefront}")
        self.session = requests.Session()
        self.session.params = {
            "country": self.storefront,
            "lang": self.language,
        }
        self.session.headers = {
            "X-Apple-Store-Front": f"{self.storefront_id} t:music31",
        }

    @functools.lru_cache()
    def get_resource(
        self,
        resource_id: str,
        entity: str = "album",
    ) -> dict:
        response = self.session.get(
            self.ITUNES_LOOKUP_API_URL,
            params={
                "id": resource_id,
                "entity": entity,
            },
        )
        try:
            response.raise_for_status()
            response_dict = response.json()
            resource = response_dict.get("results")
            assert resource
        except (
            requests.HTTPError,
            requests.exceptions.JSONDecodeError,
            AssertionError,
        ):
            AppleMusicApi._raise_response_exception(response)
        return resource

    def get_itunes_page(
        self,
        resource_type: str,
        resource_id: str,
    ) -> dict:
        response = self.session.get(
            f"{self.ITUNES_PAGE_API_URL}/{resource_type}/{resource_id}"
        )
        try:
            response.raise_for_status()
            response_dict = response.json()
            itunes_page = response_dict["storePlatformData"]["product-dv"][
                "results"
            ].get(resource_id)
            assert itunes_page
        except (
            requests.HTTPError,
            requests.exceptions.JSONDecodeError,
            AssertionError,
        ):
            AppleMusicApi._raise_response_exception(response)
        return itunes_page



from pathlib import Path

import click

from .downloader import Downloader
from tabulate import tabulate
from .enums import PostQuality


class DownloaderPost:
    QUALITY_RANK = [
        "1080pHdVideo",
        "720pHdVideo",
        "sdVideoWithPlusAudio",
        "sdVideo",
        "sd480pVideo",
        "provisionalUploadVideo",
    ]

    def __init__(
        self,
        downloader: Downloader,
        quality: PostQuality = PostQuality.BEST,
    ):
        self.downloader = downloader
        self.quality = quality

    def get_stream_url_best(self, metadata: dict) -> str:
        best_quality = next(
            (
                quality
                for quality in self.QUALITY_RANK
                if metadata["attributes"]["assetTokens"].get(quality)
            ),
            None,
        )
        return metadata["attributes"]["assetTokens"][best_quality]

    def get_stream_url_from_user(self, metadata: dict) -> str:
        qualities = list(metadata["attributes"]["assetTokens"].keys())
        table = [
            [index, quality]
            for index, quality in enumerate(
                qualities,
                start=1,
            )
        ]
        print(tabulate(table))
        choice = (
            click.prompt("Choose a quality", type=click.IntRange(1, len(table))) - 1
        )
        return metadata["attributes"]["assetTokens"][qualities[choice]]

    def get_stream_url(self, metadata: dict) -> str:
        if self.quality == PostQuality.BEST:
            stream_url = self.get_stream_url_best(metadata)
        elif self.quality == PostQuality.ASK:
            stream_url = self.get_stream_url_from_user(metadata)
        return stream_url

    def get_tags(self, metadata: dict) -> list:
        attributes = metadata["attributes"]
        return {
            "artist": attributes["artistName"],
            "date": attributes["uploadDate"],
            "title": attributes["name"],
            "title_id": int(metadata["id"]),
        }

    def get_temp_path(self, track_id: str) -> Path:
        return self.downloader.temp_path / f"{track_id}_temp.m4v"



from enum import Enum


class DownloadMode(Enum):
    YTDLP = "ytdlp"
    NM3U8DLRE = "nm3u8dlre"


class RemuxMode(Enum):
    FFMPEG = "ffmpeg"
    MP4BOX = "mp4box"


class SongCodec(Enum):
    AAC_LEGACY = "aac-legacy"
    AAC_HE_LEGACY = "aac-he-legacy"
    AAC = "aac"
    AAC_HE = "aac-he"
    AAC_BINAURAL = "aac-binaural"
    AAC_DOWNMIX = "aac-downmix"
    AAC_HE_BINAURAL = "aac-he-binaural"
    AAC_HE_DOWNMIX = "aac-he-downmix"
    ALAC = "alac"
    ATMOS = "atmos"
    ASK = "ask"


class SyncedLyricsFormat(Enum):
    LRC = "lrc"
    SRT = "srt"
    TTML = "ttml"


class MusicVideoCodec(Enum):
    H264_BEST = "h264-best"
    H265_BEST = "h265-best"
    ASK = "ask"


class PostQuality(Enum):
    BEST = "best"
    ASK = "ask"


class CoverFormat(Enum):
    JPG = "jpg"
    PNG = "png"



from __future__ import annotations

import base64
import datetime
import json
import re
import subprocess
from pathlib import Path
from xml.dom import minidom
from xml.etree import ElementTree

import click
import m3u8
from tabulate import tabulate

from .constants import SONG_CODEC_REGEX_MAP, SYNCED_LYRICS_FILE_EXTENSION_MAP
from .downloader import Downloader
from .enums import RemuxMode, SongCodec, SyncedLyricsFormat
from .models import Lyrics, StreamInfo


class DownloaderSong:
    DEFAULT_DECRYPTION_KEY = "32b8ade1769e26b1ffb8986352793fc6"
    MP4_FORMAT_CODECS = ["ec-3"]

    def __init__(
        self,
        downloader: Downloader,
        codec: SongCodec = SongCodec.AAC_LEGACY,
        synced_lyrics_format: SyncedLyricsFormat = SyncedLyricsFormat.LRC,
    ):
        self.downloader = downloader
        self.codec = codec
        self.synced_lyrics_format = synced_lyrics_format

    def get_drm_infos(self, m3u8_data: dict) -> dict:
        drm_info_raw = next(
            (
                session_data
                for session_data in m3u8_data["session_data"]
                if session_data["data_id"] == "com.apple.hls.AudioSessionKeyInfo"
            ),
            None,
        )
        if not drm_info_raw:
            return None
        return json.loads(base64.b64decode(drm_info_raw["value"]).decode("utf-8"))

    def get_asset_infos(self, m3u8_data: dict) -> dict:
        return json.loads(
            base64.b64decode(
                next(
                    session_data
                    for session_data in m3u8_data["session_data"]
                    if session_data["data_id"] == "com.apple.hls.audioAssetMetadata"
                )["value"]
            ).decode("utf-8")
        )

    def get_playlist_from_codec(self, m3u8_data: dict) -> dict | None:
        m3u8_master_playlists = [
            playlist
            for playlist in m3u8_data["playlists"]
            if re.fullmatch(
                SONG_CODEC_REGEX_MAP[self.codec], playlist["stream_info"]["audio"]
            )
        ]
        if not m3u8_master_playlists:
            return None
        m3u8_master_playlists.sort(key=lambda x: x["stream_info"]["average_bandwidth"])
        return m3u8_master_playlists[-1]

    def get_playlist_from_user(self, m3u8_data: dict) -> dict | None:
        m3u8_master_playlists = [playlist for playlist in m3u8_data["playlists"]]
        table = [
            [i, playlist["stream_info"]["audio"]]
            for i, playlist in enumerate(m3u8_master_playlists, 1)
        ]
        print(tabulate(table))
        try:
            choice = (
                click.prompt("Choose a codec", type=click.IntRange(1, len(table))) - 1
            )
        except click.exceptions.Abort:
            raise KeyboardInterrupt()
        return m3u8_master_playlists[choice]

    def get_pssh(
        self,
        drm_infos: dict,
        drm_ids: list,
    ) -> str | None:
        drm_info = next(
            (
                drm_infos[drm_id]
                for drm_id in drm_ids
                if drm_infos[drm_id].get(
                    "urn:uuid:edef8ba9-79d6-4ace-a3c8-27dcd51d21ed"
                )
                and drm_id != "1"
            ),
            None,
        )
        if not drm_info:
            return None
        return drm_info["urn:uuid:edef8ba9-79d6-4ace-a3c8-27dcd51d21ed"]["URI"]

    def get_stream_info(self, track_metadata: dict) -> StreamInfo:
        m3u8_url = track_metadata["attributes"]["extendedAssetUrls"]["enhancedHls"]
        return self._get_stream_info(m3u8_url)

    def _get_stream_info(self, m3u8_url: str) -> StreamInfo:
        stream_info = StreamInfo()
        m3u8_obj = m3u8.load(m3u8_url)
        m3u8_data = m3u8_obj.data
        drm_infos = self.get_drm_infos(m3u8_data)
        if not drm_infos:
            return stream_info
        asset_infos = self.get_asset_infos(m3u8_data)
        if self.codec == SongCodec.ASK:
            playlist = self.get_playlist_from_user(m3u8_data)
        else:
            playlist = self.get_playlist_from_codec(m3u8_data)
        if playlist is None:
            return stream_info
        stream_info.stream_url = m3u8_obj.base_uri + playlist["uri"]
        variant_id = playlist["stream_info"]["stable_variant_id"]
        drm_ids = asset_infos[variant_id]["AUDIO-SESSION-KEY-IDS"]
        pssh = self.get_pssh(drm_infos, drm_ids)
        stream_info.pssh = pssh
        stream_info.codec = playlist["stream_info"]["codecs"]
        return stream_info

    @staticmethod
    def parse_datetime_obj_from_timestamp_ttml(
        timestamp_ttml: str,
    ) -> datetime.datetime:
        mins_secs_ms = re.findall(r"\d+", timestamp_ttml)
        ms, secs, mins = 0, 0, 0
        if len(mins_secs_ms) == 2 and ":" in timestamp_ttml:
            secs, mins = int(mins_secs_ms[-1]), int(mins_secs_ms[-2])
        elif len(mins_secs_ms) == 1:
            ms = int(mins_secs_ms[-1])
        else:
            secs = float(f"{mins_secs_ms[-2]}.{mins_secs_ms[-1]}")
            if len(mins_secs_ms) > 2:
                mins = int(mins_secs_ms[-3])
        return datetime.datetime.fromtimestamp((mins * 60) + secs + (ms / 1000))

    def get_lyrics_synced_timestamp_lrc(self, timestamp_ttml: str) -> str:
        datetime_obj = self.parse_datetime_obj_from_timestamp_ttml(timestamp_ttml)
        ms_new = datetime_obj.strftime("%f")[:-3]
        if int(ms_new[-1]) >= 5:
            ms = int(f"{int(ms_new[:2]) + 1}") * 10
            datetime_obj += datetime.timedelta(milliseconds=ms) - datetime.timedelta(
                microseconds=datetime_obj.microsecond
            )
        return datetime_obj.strftime("%M:%S.%f")[:-4]

    def get_lyrics_synced_timestamp_srt(self, timestamp_ttml: str) -> str:
        datetime_obj = self.parse_datetime_obj_from_timestamp_ttml(timestamp_ttml)
        return datetime_obj.strftime("00:%M:%S,%f")[:-3]

    def get_lyrics_synced_line_lrc(self, timestamp_ttml: str, text: str) -> str:
        return f"[{self.get_lyrics_synced_timestamp_lrc(timestamp_ttml)}]{text}"

    def get_lyrics_synced_line_srt(
        self,
        index: int,
        timestamp_ttml_start: str,
        timestamp_ttml_end: str,
        text: str,
    ) -> str:
        timestamp_srt_start = self.get_lyrics_synced_timestamp_srt(timestamp_ttml_start)
        timestamp_srt_end = self.get_lyrics_synced_timestamp_srt(timestamp_ttml_end)
        return f"{index}\n{timestamp_srt_start} --> {timestamp_srt_end}\n{text}\n"

    def get_lyrics(self, track_metadata: dict) -> Lyrics:
        if not track_metadata["attributes"]["hasLyrics"]:
            return Lyrics()
        elif track_metadata.get("relationships") is None:
            track_metadata = self.downloader.apple_music_api.get_song(
                track_metadata["id"]
            )
        if track_metadata["relationships"]["lyrics"]["data"]:
            return self._get_lyrics(
                track_metadata["relationships"]["lyrics"]["data"][0]["attributes"][
                    "ttml"
                ]
            )
        else:
            return Lyrics()

    def _get_lyrics(self, lyrics_ttml: str) -> Lyrics:
        lyrics = Lyrics("", "")
        lyrics_ttml_et = ElementTree.fromstring(lyrics_ttml)
        index = 1
        for div in lyrics_ttml_et.iter("{http://www.w3.org/ns/ttml}div"):
            for p in div.iter("{http://www.w3.org/ns/ttml}p"):
                if p.text is not None:
                    lyrics.unsynced += p.text + "\n"
                if p.attrib.get("begin"):
                    if self.synced_lyrics_format == SyncedLyricsFormat.LRC:
                        lyrics.synced += f"{self.get_lyrics_synced_line_lrc(p.attrib.get('begin'), p.text)}\n"
                    elif self.synced_lyrics_format == SyncedLyricsFormat.SRT:
                        lyrics.synced += f"{self.get_lyrics_synced_line_srt(index, p.attrib.get('begin'), p.attrib.get('end'), p.text)}\n"
                    elif self.synced_lyrics_format == SyncedLyricsFormat.TTML:
                        if not lyrics.synced:
                            lyrics.synced = minidom.parseString(
                                lyrics_ttml
                            ).toprettyxml()
                        continue
                    lyrics.synced += "\n"
                    index += 1
            lyrics.unsynced += "\n"
        lyrics.unsynced = lyrics.unsynced[:-2]
        return lyrics

    def get_tags(self, webplayback: dict, lyrics_unsynced: str) -> dict:
        tags_raw = webplayback["assets"][0]["metadata"]
        tags = {
            "album": tags_raw["playlistName"],
            "album_artist": tags_raw["playlistArtistName"],
            "album_id": int(tags_raw["playlistId"]),
            "album_sort": tags_raw["sort-album"],
            "artist": tags_raw["artistName"],
            "artist_id": int(tags_raw["artistId"]),
            "artist_sort": tags_raw["sort-artist"],
            "comments": tags_raw.get("comments"),
            "compilation": tags_raw["compilation"],
            "composer": tags_raw.get("composerName"),
            "composer_id": (
                int(tags_raw.get("composerId")) if tags_raw.get("composerId") else None
            ),
            "composer_sort": tags_raw.get("sort-composer"),
            "copyright": tags_raw.get("copyright"),
            "date": (
                self.downloader.sanitize_date(tags_raw["releaseDate"])
                if tags_raw.get("releaseDate")
                else None
            ),
            "disc": tags_raw["discNumber"],
            "disc_total": tags_raw["discCount"],
            "gapless": tags_raw["gapless"],
            "genre": tags_raw["genre"],
            "genre_id": tags_raw["genreId"],
            "lyrics": lyrics_unsynced if lyrics_unsynced else None,
            "media_type": 1,
            "rating": tags_raw["explicit"],
            "storefront": tags_raw["s"],
            "title": tags_raw["itemName"],
            "title_id": int(tags_raw["itemId"]),
            "title_sort": tags_raw["sort-name"],
            "track": tags_raw["trackNumber"],
            "track_total": tags_raw["trackCount"],
            "xid": tags_raw.get("xid"),
        }
        return tags

    def get_encrypted_path(self, track_id: str) -> Path:
        return self.downloader.temp_path / f"{track_id}_encrypted.m4a"

    def get_decrypted_path(self, track_id: str) -> Path:
        return self.downloader.temp_path / f"{track_id}_decrypted.m4a"

    def get_remuxed_path(self, track_id: str) -> Path:
        return self.downloader.temp_path / f"{track_id}_remuxed.m4a"

    def fix_key_id(self, encrypted_path: Path):
        count = 0
        with open(encrypted_path, "rb+") as file:
            while data := file.read(4096):
                pos = file.tell()
                i = 0
                while tenc := max(0, data.find(b"tenc", i)):
                    kid = tenc + 12
                    file.seek(max(0, pos - 4096) + kid, 0)
                    file.write(bytes.fromhex(f"{count:032}"))
                    count += 1
                    i = kid + 1
                file.seek(pos, 0)

    def decrypt(
        self,
        encrypted_path: Path,
        decrypted_path: Path,
        decryption_key: str,
    ):
        self.fix_key_id(encrypted_path)
        subprocess.run(
            [
                self.downloader.mp4decrypt_path_full,
                encrypted_path,
                "--key",
                f"00000000000000000000000000000001:{decryption_key}",
                "--key",
                f"00000000000000000000000000000000:{self.DEFAULT_DECRYPTION_KEY}",
                decrypted_path,
            ],
            check=True,
        )

    def remux(self, decrypted_path: Path, remuxed_path: Path, codec: str):
        if self.downloader.remux_mode == RemuxMode.MP4BOX:
            self.remux_mp4box(decrypted_path, remuxed_path)
        elif self.downloader.remux_mode == RemuxMode.FFMPEG:
            self.remux_ffmpeg(decrypted_path, remuxed_path, codec)

    def remux_mp4box(self, decrypted_path: Path, remuxed_path: Path):
        subprocess.run(
            [
                self.downloader.mp4box_path_full,
                "-quiet",
                "-add",
                decrypted_path,
                "-itags",
                "artist=placeholder",
                "-new",
                remuxed_path,
            ],
            check=True,
        )

    def remux_ffmpeg(
        self,
        decrypted_path: Path,
        remuxed_path: Path,
        codec: str,
    ):
        use_mp4_format = any(
            codec.startswith(possible_codec)
            for possible_codec in self.MP4_FORMAT_CODECS
        )
        subprocess.run(
            [
                self.downloader.ffmpeg_path_full,
                "-loglevel",
                "error",
                "-y",
                "-i",
                decrypted_path,
                "-c",
                "copy",
                "-f",
                "mp4" if use_mp4_format else "ipod",
                "-movflags",
                "+faststart",
                remuxed_path,
            ],
            check=True,
        )

    def get_lyrics_synced_path(self, final_path: Path) -> Path:
        return final_path.with_suffix(
            SYNCED_LYRICS_FILE_EXTENSION_MAP[self.synced_lyrics_format]
        )

    def get_cover_path(self, final_path: Path) -> Path:
        return final_path.parent / f"Cover.{self.downloader.cover_format.value}"

    def save_lyrics_synced(self, lyrics_synced_path: Path, lyrics_synced: str):
        lyrics_synced_path.parent.mkdir(parents=True, exist_ok=True)
        lyrics_synced_path.write_text(lyrics_synced, encoding="utf8")



import base64
import subprocess
from pathlib import Path

import m3u8
from pywidevine import PSSH
from pywidevine.license_protocol_pb2 import WidevinePsshData

from .downloader_song import DownloaderSong
from .enums import RemuxMode, SongCodec
from .models import StreamInfo


class DownloaderSongLegacy(DownloaderSong):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_stream_info(self, webplayback: dict) -> StreamInfo:
        flavor = "32:ctrp64" if self.codec == SongCodec.AAC_HE_LEGACY else "28:ctrp256"
        stream_info = StreamInfo()
        stream_info.stream_url = next(
            i for i in webplayback["assets"] if i["flavor"] == flavor
        )["URL"]
        m3u8_obj = m3u8.load(stream_info.stream_url)
        stream_info.pssh = m3u8_obj.keys[0].uri
        return stream_info

    def get_decryption_key(self, pssh: str, track_id: str) -> str:
        widevine_pssh_data = WidevinePsshData()
        widevine_pssh_data.algorithm = 1
        widevine_pssh_data.key_ids.append(base64.b64decode(pssh.split(",")[1]))
        pssh_obj = PSSH(widevine_pssh_data.SerializeToString())
        cdm_session = self.downloader.cdm.open()
        challenge = base64.b64encode(
            self.downloader.cdm.get_license_challenge(cdm_session, pssh_obj)
        ).decode()
        license = self.downloader.apple_music_api.get_widevine_license(
            track_id,
            pssh,
            challenge,
        )
        self.downloader.cdm.parse_license(cdm_session, license)
        decryption_key = next(
            i for i in self.downloader.cdm.get_keys(cdm_session) if i.type == "CONTENT"
        ).key.hex()
        self.downloader.cdm.close(cdm_session)
        return decryption_key

    def decrypt(
        self,
        encrypted_path: Path,
        decrypted_path: Path,
        decryption_key: str,
    ):
        self.fix_key_id(encrypted_path)
        subprocess.run(
            [
                self.downloader.mp4decrypt_path_full,
                encrypted_path,
                "--key",
                f"1:{decryption_key}",
                decrypted_path,
            ],
            check=True,
        )

    def remux_mp4box(self, decrypted_path: Path, remuxed_path: Path) -> None:
        subprocess.run(
            [
                self.downloader.mp4box_path_full,
                "-quiet",
                "-add",
                decrypted_path,
                "-itags",
                "artist=placeholder",
                "-new",
                remuxed_path,
            ],
            check=True,
        )

    def remux_ffmpeg(
        self,
        decryption_key: str,
        encrypted_path: Path,
        remuxed_path: Path,
    ):
        subprocess.run(
            [
                self.downloader.ffmpeg_path_full,
                "-loglevel",
                "error",
                "-y",
                "-decryption_key",
                decryption_key,
                "-i",
                encrypted_path,
                "-c",
                "copy",
                "-movflags",
                "+faststart",
                remuxed_path,
            ],
            check=True,
        )

    def remux(
        self,
        encrypted_path: Path,
        decrypted_path: Path,
        remuxed_path: Path,
        decryption_key: str,
    ):
        if self.downloader.remux_mode == RemuxMode.FFMPEG:
            self.remux_ffmpeg(decryption_key, encrypted_path, remuxed_path)
        elif self.downloader.remux_mode == RemuxMode.MP4BOX:
            self.decrypt(encrypted_path, decrypted_path, decryption_key)
            self.remux_mp4box(decrypted_path, remuxed_path)
